/*
      Copyright (C) 2009 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */

using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;
using BEPUphysics;
using BEPUphysics.Entities;
using BEPUphysics.DataStructures;

namespace BasicSetupDemo
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class BowlingGame : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        /// <summary>
        /// Handles and responds to user input.
        /// </summary>
        public User user;
        /// <summary>
        /// World in which the simulation runs.
        /// </summary>
        public Space space;
        /// <summary>
        /// Model to use for the bowling balls.  Used when spawning more balls.
        /// </summary>
        public Model bowlingBallModel;
#if XBOX360
        /// <summary>
        /// Contains the latest snapshot of the gamepad's input state.
        /// </summary>
        public GamePadState gamePadState;
#else
        /// <summary>
        /// Contains the latest snapshot of the keyboard's input state.
        /// </summary>
        public KeyboardState keyboardState;
        /// <summary>
        /// Contains the latest snapshot of the mouse's input state.
        /// </summary>
        public MouseState mouseState;
#endif


        public BowlingGame()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here

            //The 'user' component collects and uses input.
            user = new User(new Camera(new Vector3(0, 3, 10), 0, 0, 10), this);
            Components.Add(user);

            

            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // TODO: use this.Content to load your game content here

            //This 1x1x1 cube model will represent the box entities in the space.
            Model cubeModel = Content.Load<Model>("cube");

            Model pinModel = Content.Load<Model>("bowlingpin");
            Model alleyModel = Content.Load<Model>("bowlingalley");
            bowlingBallModel = Content.Load<Model>("sphere");

            //Construct a new space for the physics simulation to occur within.
            space = new Space();

            //Set the gravity of the simulation by accessing the simulation settings of the space.
            //It defaults to (0,0,0); this changes it to an 'earth like' gravity.
            //Try looking around in the space's simulationSettings to familiarize yourself with the various options.
            space.simulationSettings.motionUpdate.gravity = new Vector3(0, -9.81f * 1.5f, 0);

            //Enable the fancier rotation integration.  This isn't really necessary, but you can turn it off and on to see the (fairly small) difference.
            space.simulationSettings.motionUpdate.conserveAngularMomentum = true;
            space.simulationSettings.motionUpdate.useRK4AngularIntegration = true;

            //Make a box representing the ground and add it to the space.
            //The Box is an "Entity," the main simulation object type of BEPUphysics.
            //Examples of other entities include cones, spheres, cylinders, and a bunch more (a full listing is in the BEPUphysics.Entities namespace).

            //Every entity has a set of constructors.  Some have a parameter for mass, others don't.
            //Constructors that allow the user to specify a mass create 'dynamic' entiites which fall, bounce around, and generally work like expected.
            //Constructors that have no mass parameter create a create 'kinematic' entities.  These can be thought of as having infinite mass.
            //This box being added is representing the ground, so the width and length are extended and it is kinematic.
            space.add(new Box(new Vector3(0,-.5f, 0), 10, 1f, 50));

            int[] indices;
            //Instead of a box for a bowling alley, we could also use a static triangle group.
            /*
            StaticTriangleGroup.StaticTriangleGroupVertex[] staticTriangleVertices;
            
            StaticTriangleGroup.getVerticesAndIndicesFromModel(alleyModel, out staticTriangleVertices, out indices);
            //Note the margin is 0.  Usually this isn't the case, but in some cases it can help reduce the appearance of
            //bumping when rolling over triangle transitions.
            TriangleMesh mesh = new TriangleMesh(staticTriangleVertices, indices, 0);
            mesh.worldMatrix = Matrix.CreateTranslation(new Vector3(0, -.24f, 0));
            StaticTriangleGroup group = new StaticTriangleGroup(mesh);
            space.add(group);
            Components.Add(new StaticModel(alleyModel, group.worldMatrix, this));
            */

            //Extract vertices from the model for the pin.
            Vector3[] vertices;
            TriangleMesh.getVerticesAndIndicesFromModel(pinModel, out vertices, out indices);
            
            
            //How about a game of tenpin?

            List<Vector3> points = new List<Vector3>();
            List<int> triangleIndices = new List<int>();
            //Technically, we don't need to run the convex hullifier on this pointset,
            //since the entity constructor would deal with it.  However, doing a first pass
            //will reduce the number of points being dealt with later on.
            Toolbox.getConvexHull(new List<Vector3>(vertices), triangleIndices, points);
            //Create the pins.
            //Note that the center positions of the entities are the centers of volume, so
            //to get them sitting right where you want, you'll need to figure out the difference
            //between the origin of your model and the center position of the created convex hull.
            //Comparing the model's origin to that of a ConvexHull created from its pointset
            //will provide the offset.  That offset can then be used to precisely position the hulls.
            //In the following code, that offset has been taken into account in the initial positions,
            //and is used to translate the graphical model of the pin to the right spot.
            Matrix translation = Matrix.CreateTranslation(new Vector3(0, -.55f, 0));
            ConvexHull pin = new ConvexHull(new Vector3(0, .5f, -14.5f), points, 1);
            space.add(pin);
            Components.Add(new EntityModel(pin, pinModel, translation, this));

            for (int k = 0; k < 4; k++)
            {
                pin = new ConvexHull(new Vector3(-1.5f + k * 1, .5f, -19), points, 1);
                space.add(pin);
                Components.Add(new EntityModel(pin, pinModel, translation, this));
            }
            for (int k = 0; k < 3; k++)
            {
                pin = new ConvexHull(new Vector3(-1f + k * 1, .5f, -17.5f), points, 1);
                space.add(pin);
                Components.Add(new EntityModel(pin, pinModel, translation, this));
            }
            for (int k = 0; k < 2; k++)
            {
                pin = new ConvexHull(new Vector3(-.5f + k * 1, .5f, -16), points, 1);
                space.add(pin);
                Components.Add(new EntityModel(pin, pinModel, translation, this));
            }

            //Go through the list of entities in the space and create a graphical representation for them.
            foreach (Entity e in space.entities)
            {
                Box box = e as Box;
                if (box != null) //This won't create any graphics for an entity that isn't a box since the model being used is a box.
                {
                    
                    Matrix scaling = Matrix.CreateScale(box.width, box.height, box.length); //Since the cube model is 1x1x1, it needs to be scaled to match the size of each individual box.
                    //Add the drawable game component for this entity to the game.
                    Components.Add(new EntityModel(e, cubeModel, scaling, this));
                }
            }

        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
#if XBOX360
            gamePadState = GamePad.GetState(0);
#else
            keyboardState = Keyboard.GetState();
            mouseState = Mouse.GetState();
#endif
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed
#if XBOX360
                )
#else
                || keyboardState.IsKeyDown(Keys.Escape))
#endif
                this.Exit();

            // TODO: Add your update logic here

            //Steps the simulation forward one time step.
            space.update(gameTime);

            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            // TODO: Add your drawing code here

            base.Draw(gameTime);
        }
    }
}
